#!/usr/bin/env python2.3
################################################################################
#
#       This file is part of the GQL (Graphical Query Language) Toolkit
#
#       file:   GQL.py
#       author: Alexander Schliep (alexander@schliep.org)
#
#       Copyright (C) 2003-2004 Alexander Schliep
#
#       Contact: alexander@schliep.org
#
#       Information: http://ghmm.org/gql
#
#	GQL is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; either version 2 of the License, or
#	(at your option) any later version.
#
#	GQL is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with GQL; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#
#       This file is version $Revision: 1.22 $
#                       from $Date: 2004/12/17 14:38:19 $
#             last change by $Author: filho $.
#
################################################################################
#
#
#
#----- Globals -----------------------------------------------------------------
gGQLVersion = 0.6
gGQLBuilddate = "1/15/2003"
gNCBIURL = "http://www.ncbi.nlm.nih.gov/entrez/query.fcgi?cmd=search&db=nucleotide&term=%s[accn]"
MISSING_DATA = -9999.99

#-------------------------------------------------------------------------------
from Tkinter import *
import tkFont
import webbrowser
from tkFileDialog import askopenfilename, asksaveasfilename
from tkMessageBox import askokcancel, showerror, askyesno
import tkSimpleDialog
#from ScrolledText import *
import os
import string
import numpy as Numeric
from GQLQuery import GQLQuery
from GQLQueryEditor import *
from GQLUtil import extension, stripPath
from GQLGui import WMExtrasGeometry, ImageCache, AboutBox, SplashScreen, \
     ScrolledText, ResultsBox, ProfileCanvas, ClassesColorScheme
from ghmm import SequenceSet, Float
import GQLIcons

# missing data value

def minWithoutMissing(array):
    result = []
    for x in array:
        if x != MISSING_DATA: # simbol for missing data
           result.append(x)
    return min(result)

class ProfileSet:
    """ Class to hold all information pertinent to a profile data set

        - profiles are referenced via an integer (from 0 to #profiles - 1)
        - all additional info is stored in vectors
        - id to index dictionary is used for cross-referencing

    """

    def __init__(self):
        self.new()

    def new(self):
        self.virgin = True

        self.size = 0

        self.profile = []
        self.info = []
        self.acc = []
        self.genename = []
        self.outlink = []
        self.likelihood = []
	self.seq_classes  = [] # classes of the sequences (when they exists)
 	self.classes_no = 0 # number of classes
	self.max_class = 0 # class with maximun value
	self.classes = []

        self.xrange = [0,0]
        self.yrange = [1E300,0]
        self.likelihoodrange = [0,0]

        self.ghmm_seqs = None # a SequenceSet object.
        self.id2Index = {}
        self.fileName = None
        self.missingRate = 0.0

    def beforeRead(self):
        if not self.virgin:
            del(self.profile)

            del(self.acc)
            del(self.genename)
            del(self.outlink)
            del(self.info)
	    del(self.seq_classes)
	    del(self.classes_no)
	    del(self.classes)

            del(self.likelihood)
            del(self.id2Index)
            #XXX del(self.ghmm_seqs.free())
            self.new()

    def afterRead(self):
        pass

    def ReadDataFromCaged(self, fileName):
        self.fileName = fileName
        self.beforeRead()

        #--- create the SQD file
        # File format of caged file (and a Matlab compatible variant)
        # 1). line: Textual description "Gene name\tAccession" for cage
        # Genname Acc# value1 ... valueN (tab-delim)

        file = open(fileName,'r')
        lines = file.readlines()
        #print "ReadDataFromCaged. File has %d lines" % len(lines)
        file.close()

	classes = [] # keep track of the existing classes

        outfile = open(fileName + ".sqd",'w')
        outfile.write("SEQD = {\n      O = {\n")

        self.headerLine = lines[0]
        if string.find(self.headerLine, "Gene") >= 0:
            start_i = 1
        else:
            print "ReadDataFromCaged: No Header, assuming Matlab compat"
            start_i = 0

        def replaceMissing(x):
            # XXX Fix me !
            if x == '':
                return MISSING_DATA
#                 return 0.0
            elif x == '_M': # 'official' missing value indicator
                return MISSING_DATA
#                 return 0.0
            elif x == 'NA': # 'official' missing value indicator
                return MISSING_DATA
#                 return 0.0
            else:
                return x

        nrTimeSteps = len(lines[0].split('\t')) - 2

        i = 1 # We count the profiles from 1
        for l in lines[start_i:]:
            items = l.split('\t')
            items[-1] = string.strip(items[-1],'\r\n')

            # Fix too-short sequences: values missing at the end and no tabs
            if len(items[2:]) < nrTimeSteps:
                items += ([''] * (nrTimeSteps - len(items[2:])))

            values = map(replaceMissing, items[2:])

            if MISSING_DATA != max(map(float,values)): # data cotains not only missinda data

                self.genename.append(string.strip(str(items[0])))
                self.acc.append(string.strip(str(items[0])))

                # try to read label information ...
	        try:
	    	    class_aux = int(string.strip(str(items[1])));
                    self.seq_classes.append(class_aux) # if it is only a number in the commentar, asume it is the class of the sequence
		    if class_aux not in self.classes:
			self.classes.append(class_aux)
	        except ValueError:
                    self.seq_classes.append(-1) # no class information ...

                outfile.write("(%d) |   1| " % i)
                outfile.write(string.join(map(str,values), ', '))
                outfile.write(', 9999.99;\n')
                self.outlink.append("www.somedb.org/%d.html" % i)
                i += 1


	self.classes_no = len(self.classes)
	if(self.classes == []):
		self.max_class = 0
	else:
		self.max_class = max(1,max(self.classes))

	outfile.write("};};")
        outfile.close()

        self.ghmm_seqs = SequenceSet(Float(), fileName + ".sqd")

        for i in xrange(len(self.ghmm_seqs)):
            p = list(self.ghmm_seqs[i])
            # NOTE: Sqd files are already tagged with 9999 value as end marker
            self.addProfile("#%d" % i, p[:-1], "info about %d" % i)
            #self.acc.append("acc%d" % i)
            #self.genename.append("genename%d" % i)
            #self.outlink.append("www.somedb.org/%d.html" % i)

        self.afterRead()
        self.virgin = False
        self.missingDataRate()
        return self.ghmm_seqs

    def ReadDataFromDSequences(self, ghmm_seqs,fileName):
        print "read data from ds"
        self.fileName = fileName
        self.beforeRead()
        self.ghmm_seqs = ghmm_seqs
        for i in xrange(len(ghmm_seqs)):
            p = list(ghmm_seqs[i])
            # NOTE: Sqd files are already tagged with 9999 value as end marker
            self.addProfile("#%d" % i, p[:-1], "info about %d" % i)
            self.acc.append("acc%d" % i)
            self.genename.append("genename%d" % i)
            self.outlink.append("www.somedb.org/%d.html" % i)
	    try:
		class_aux = int(ghmm_seqs.getSeqLabel(i))
		self.seq_classes.append(class_aux)
		if class_aux not in self.classes:
			self.classes.append(class_aux)
            except ValueError:
               self.seq_classes.append(-1) # no class information ...

	if(self.classes == []):
		self.max_class = 0
	else:
		self.max_class = max(1,max(self.classes))
	self.classes_no = len(self.classes)
        self.missingDataRate()
        self.afterRead()
        self.virgin = False

    def missingDataRate(self):
        missingCount = 0.0
        totalCount = len(self.ghmm_seqs)*(len(self.ghmm_seqs[0])-1)
        for sequence in self.ghmm_seqs:
            for value in sequence:
                if value == MISSING_DATA:
                    missingCount += 1.0
        self.missingRate = missingCount/totalCount
        print "Missing data rate %f"%self.missingRate
        

    def addProfile(self, id, profile, info,classe=-1):
        # Determine min and max of the y coordinate
        self.yrange[0] = min(self.yrange[0], minWithoutMissing(profile))
        self.yrange[1] = max(self.yrange[1], max(profile))


        # Determine maximal x-value!
        self.xrange[1] = max(self.xrange[1], len(profile)-1)

        self.profile.append(profile)
        self.info.append("This is profile id = %s" % id)
        self.id2Index[id] = self.size
        self.size += 1
        if (classe != -1):
	    self.seq_classes.append(classe)

    def __getitem__(self, i):
        return self.profile[i]

    def Info(self,i):
        return "%s: Acc#%s, its genename is %s" % (self.info[i], self.acc[i], self.genename[i])

    def __len__(self):
        return self.size

    def GHMMProfileSet(self): # XXX Should be GHMMSequenceSet
        return self.ghmm_seqs

    def getclass(self, i):
        return self.seq_classes[i]

    def tabDelim(self,query,delim='\t'):
        """Return id acc genename timepoint1 ... timepointn in tab-delimeted format"""
        result = ""

        for i in query.Result():
            result += "%s%s%2.2f%s%s%s%s%s" % (str(i), delim,
                                            query.Likelihood(i), delim,
                                            self.acc[i], delim,
                                            self.genename[i], delim)
            result += string.join(map(str,self.profile[i][:-1]),delim) # DOnt want final 9999.9
            result += "\n"

        return result

    def getSubset(self,ids):
	subset = ProfileSet()
	subset.virgin = False
    	for i in ids:
            subset.genename.append(self.genename[i])
            subset.acc.append(self.acc[i])
            subset.outlink.append(self.outlink[i])
	    subset.addProfile("#%d" % i, self.profile[i], "info about %d" % i)
            subset.seq_classes.append(self.seq_classes[i])

        subset.classes_no = self.classes_no
        subset.max_class = self.max_class
	subset.classes = self.classes
	subset.fileName = self.fileName
        subset.ghmm_seqs = self.ghmm_seqs.getSubset(ids)
	return subset


class ProfileDisplay(Frame):

    def __init__(self, no_states, parent = None):
        self.canvas_width = 700
        self.canvas_height = 500
	self.no_states=no_states
        Frame.__init__(self, parent, relief=GROOVE, bd=2)
        self.makeWidgets()
        self.profileSet = None
        self.parent = parent
        #self.lastRank = int(self.likelihoodCutoff.get())


    def makeWidgets(self):
        label = Label(self, text="Query Result", anchor=W)
        label.pack(anchor=W, side=TOP, padx = 4)

        sunkenFrame = Frame(self, relief=SUNKEN, bd=2)
        self.canvas = ProfileCanvas(self, sunkenFrame, self.canvas_width,
                                    self.canvas_height)

        self.canvas.pack(anchor=W, side=TOP, expand=0, padx=4, pady=4)
        sunkenFrame.pack(anchor=W, side=TOP, expand=0, padx=4, pady=4)

        self.framePath = Frame(self,relief=RIDGE, bd=2,height=15,width =self.canvas_width)

	self.canvasPath = Canvas(self.framePath,height=15,width=self.canvas_width)
	self.canvasPath.pack(side=LEFT, fill=X, expand=NO)
	self.canvasPathItens = []

	#self.btn1 = Button(self.framePath, text="", anchor=W, default=DISABLED)
        #self.btn1.pack(anchor=W, side=BOTTOM, expand=1, fill=X)
        self.framePath.pack(anchor=W, side=TOP, expand=1, fill=X, padx = 4, pady=0)

	self.likelihoodCutoff = IntVar()
        self.likelihoodScale = Scale(self, command = self.updateLikelihood,
                                     from_ = 4000, to = 1, resolution = 1,
                                     orient = HORIZONTAL, relief = FLAT,
                                     showvalue = 1,
                                     #tickinterval = 0.1,
                                     variable = self.likelihoodCutoff,
                                     length = 300, label="Similarity Rank")
        self.likelihoodScale.pack(anchor=E, side=TOP, expand=0, padx=4, pady=4)

        frame = Frame(self,relief=RIDGE, bd=2)
        self.profileInfo = Label(frame, text="", anchor=W)
        self.profileInfo.pack(anchor=W, side=BOTTOM, expand=1, fill=X)
        frame.pack(anchor=W, side=BOTTOM, expand=1, fill=X, padx = 4, pady=4)


    def DisplayProfileSet(self, set):
        if self.profileSet is not None:
            self.canvas.Clear()
        self.profileSet = set
        self.likelihoodScale.config(from_=len(set))
        self.likelihoodScale.set(len(set))
        self.canvas.DisplayProfileSet(set)
        self.parent.queryEditor.configureRange(self.canvas.yrange[0],self.canvas.yrange[1])
	self.no_states = self.parent.query.queryLength


    def ClearInfo(self):
        self.profileInfo.config(text="")
	self.HidePathSlider()

    def ProfileInfo(self,the_text,profile):
        self.profileInfo.config(text=the_text)
	# check is the query was modified (the viterbi paths are recalculated in this case)
	self.parent.query.isModified()
	self.UpdatePathSlider(profile)

    def UpdatePathSlider(self,profile_id):
        path = self.parent.query.viterbiPaths[0][profile_id]
    	pathLen = len(path) - 1 # ignoring the last state and the space not avaliable at the graph
	print "viterbi", path
	statesFrequency = Numeric.zeros(self.no_states,Numeric.Float)

	for state in path[:-1]: # the end state is not drawn
            statesFrequency[state] +=1.0


        statelenght = self.canvas_width/(pathLen-1.0)
	# the first state have a smaler size (only half the original size)
	actualAux = 0
	previousX = (statesFrequency[0]-0.5)*(statelenght)
        self.canvasPath.create_rectangle(0,0,previousX,15,
	                                 outline=ClassesColorScheme(0,self.no_states),fill=ClassesColorScheme(0,self.no_states))
        i = 1

	for state in statesFrequency[1:-1]:
	    actualX = previousX + (statesFrequency[i])*statelenght
	    self.canvasPath.create_rectangle(previousX,0,actualX,15,outline=ClassesColorScheme(i,self.no_states),
	                                     fill=ClassesColorScheme(i,self.no_states))
	    previousX = actualX
	    i +=1

        self.canvasPath.create_rectangle(previousX,0,700,15,
	                                 outline=ClassesColorScheme(self.no_states-1,self.no_states),
					 fill=ClassesColorScheme(self.no_states-1,self.no_states))



    def HidePathSlider(self):
    	self.canvasPath.delete(ALL)

    def Update(self):
        # Only called for new data set
        self.likelihoodScale.config(from_=len(self.profileSet))

    def update(self,showProfiles,hideProfiles):
        if self.no_states != self.parent.query.queryLength:
		self.no_states = self.parent.query.queryLength
	self.canvasPath.delete(ALL)
        self.canvas.ProfileUpdate(showProfiles,hideProfiles)
        l = self.parent.query.LowestRankedLikelihood()
        self.likelihoodScale.config(label="Similarity Rank   (log-likelihood exceeds %3.2f)" % l)

    def updateLikelihood(self, newRank):
        self.parent.query.setRankCutoff(int(newRank))
        l = self.parent.query.LowestRankedLikelihood()
        self.likelihoodScale.config(label="Similarity Rank   (log-likelihood exceeds %3.2f)" % l)


class GQLApp(Frame):

    def __init__(self, parent=None):
        Frame.__init__(self,parent)
        GQLIcons.Init()
        Splash = SplashScreen(self.master)

        #------- App init ---------------------------------------------
        self.pack(expand = 0)
        self.makeMenuBar()
        self.queryEditor = GQLQueryEditor(3,self)
        self.queryEditor.pack(anchor=N, side=LEFT, padx=4, pady=4, expand=0)

        self.profileDisplay = ProfileDisplay(3,self)
        self.profileDisplay.pack(anchor=N, side=RIGHT, padx=4, pady=4, expand=0)
        self.master.title("GQL")

        # Containers
        self.profileSet = ProfileSet() # We keep one ...
        self.ghmmSeqSet = None
        self.query = GQLQuery()
        self.queryEditor.overrideQuery(self.query)

        self.query.addListener(self.profileDisplay)

        #------- Tk vodoo ---------------------------------------------
        Splash.Destroy()

        # Fix focus and stacking
        if os.name == 'nt' or os.name == 'dos':
            self.master.tkraise()
            self.master.focus_force()
        else:
            self.tkraise()

        # Make AlgoWins requested size its minimal size to keep
        # toolbar from vanishing when changing window size
        # Packer has been running due to splash screen
        wmExtras = WMExtrasGeometry(self.master)
        if os.name == 'nt' or os.name == 'dos':
            self.master.minsize(self.master.winfo_reqwidth(),
                                self.master.winfo_reqheight() +
                                wmExtras[1])
        else: # Unix & Mac
            self.master.minsize(self.master.winfo_reqwidth(),
                                self.master.winfo_reqheight() +
                                wmExtras[0] + wmExtras[1])


    ############################################################
    #
    # Create GUI
    #
    def makeMenuBar(self):
        """ *Internal* Now using Tk 8.0 style menues """
        self.menubar = Menu(self, tearoff=0)

        # Add file menu
        self.fileMenu = Menu(self.menubar, tearoff=0)
        self.fileMenu.add_command(label='Open Data Set...',
                                  command=self.OpenDataSet)
        self.fileMenu.add_separator()
        self.fileMenu.add_command(label='Open Query...',
                                  command=self.OpenQuery)
        self.fileMenu.add_command(label='New Query...',
                                  command=self.NewQuery)
        self.fileMenu.add_command(label='Save Query...',
                                  command=self.SaveQuery)
        self.fileMenu.add_separator()
        #self.fileMenu.add_command(label='Preferences...',
        #                          command=self.Preferences)
        #self.fileMenu.add_separator()
        self.fileMenu.add_command(label='Export Result as EPS...',
                                  command=self.ExportEPSF)
        self.fileMenu.add_command(label='Show Result...',
                                  command=self.ShowResult)
        self.fileMenu.add_command(label='Save Result...',
                                  command=self.SaveResult)
        self.fileMenu.add_separator()
        self.fileMenu.add_command(label='Quit',
                                  command=self.Quit)
        self.menubar.add_cascade(label="File", menu=self.fileMenu,
                                 underline=0)


        # On a Mac we put our about box under the Apple menu ...
        if os.name == 'mac':
            self.apple=Menu(self.menubar, tearoff=0, name='apple')
            self.apple.add_command(label='About GQL',
                                   command=self.AboutBox)
            self.apple.add_command(label='Help',
                                   command=self.HelpBox)
            self.menubar.add_cascade(menu=self.apple)

        else: # ... on other systems we add a help menu
            self.helpMenu=Menu(self.menubar, tearoff=0, name='help')
            self.helpMenu.add_command(label='About GQL',
                                      command=self.AboutBox)
            self.helpMenu.add_command(label='Help',
                                   command=self.HelpBox)
            self.menubar.add_cascade(label="Help", menu=self.helpMenu,
                                     underline=0)

        self.master.configure(menu=self.menubar)


    ############################################################
    #
    # Menu call backs
    #
    def OpenDataSet(self):
        fileName = askopenfilename(title="Open Expression Data Set ",
                                   defaultextension=".txt",
                                   filetypes = [("Cage compatible", ".txt"),
                                                ("GHMM SQD File", ".sqd")]
                                   )
        if fileName != '':

            if self.ghmmSeqSet is not None:
                del(self.ghmmSeqSet)
            if extension(fileName) == 'sqd': # READ from SQD
                self.ghmmSeqSet = SequenceSet(Float(),fileName)
                self.profileSet.ReadDataFromDSequences(self.ghmmSeqSet,fileName)
            elif extension(fileName) == 'txt': # READ from CAGED
                self.ghmmSeqSet = self.profileSet.ReadDataFromCaged(fileName)
            else:
                print "GQL::OpenDataSet: invalid file type '%s'" % fileName

            self.profileDisplay.DisplayProfileSet(self.profileSet)
            self.query.setData(self.profileSet)



    def OpenQuery(self):
        fileName = askopenfilename(title="Open Query",
                                   defaultextension=".smo",
                                   filetypes = [("GHMM SModel File",".smo")]
                                   )

        if fileName != '':
            self.query.openQuery(fileName)
            self.queryEditor.fromQuery(self.query)
            self.query.runQuery()


    def NewQuery(self):
        """ Reset all values: Currently this does maintain the number of steps """
        queryLength = self.query.queryLength
        # XXX Get default values elsewhere (config file?)
        mean = [0.0] * queryLength
        variance = [0.5] * queryLength
        duration = [1.1] * queryLength
        self.query.newQuery(queryLength, mean, variance, duration, 0)
        self.queryEditor.fromQuery(self.query)

    def SaveQuery(self):
        fileName = asksaveasfilename(title="Save Query",
                                     defaultextension=".smo",
                                     filetypes = [("GHMM SModel File",".smo")]
                                     )
        if fileName != '':
            self.query.saveQuery(fileName)
        

    def ExportEPSF(self,baseName=None):
        if baseName == None:
            fileName = asksaveasfilename(title="Export EPSF",
                                         defaultextension=".eps",
                                         filetypes = [  ("Encapsulated PS", ".eps")
                                                        ,("Postscript", ".ps")
                                                        ]
                                         )
        else:
            fileName = baseName + ".eps"

        if fileName != '': 
            """ Produce an EPSF of canvas in fileName. Note: Graph gets scaled
            and rotated as to maximize size while still fitting on paper """ 
            bb = self.profileDisplay.canvas.bbox("all") # Bounding box of all elements on canvas
            # Give 10 pixels room to breathe
            x = max(bb[0] - 10,0)
            y = max(bb[1] - 10,0)
            width=bb[2] - bb[0] + 10
            height=bb[3] - bb[1] + 10

            printablePageHeight=280 #mm
            printablePageWidth =190 #mm

            printableRatio=printablePageHeight/printablePageWidth
        
            bbRatio = height/width

            if bbRatio > printableRatio: # Height gives limiting dimension
                self.profileDisplay.canvas.postscript(file=fileName,
                                                      pageheight="%dm" % printablePageHeight,
                                                      x=x,y=y,height=height,width=width)
            else:
                self.profileDisplay.canvas.postscript(file=fileName,
                                                      pagewidth="%dm" % printablePageWidth,
                                                      x=x,y=y,height=height,width=width)

    def ShowResult(self):
        text = self.profileSet.tabDelim(self.query)
        line2acc = []
        for i in self.query.Result():
            line2acc.append(self.profileSet.acc[i])
        d = ResultsBox(self,text,line2acc)


    def SaveResult(self,baseName=None):
        if baseName == None:
            fileName = asksaveasfilename(title="Save Result as plain text",
                                         defaultextension=".txt",
                                         filetypes = [("Text", ".txt")]
                                         )
        else:
            fileName = baseName + ".txt"
        if fileName != '':
            file = open(fileName, 'w')
            file.write(self.profileSet.tabDelim(self.query))
            file.close()            
    
    def Preferences(self):
        pass

    def Quit(self):
        if askokcancel("Quit","Do you really want to quit?"):
            Frame.quit(self)
            #self.CleanUp()

    def AboutBox(self):
        d = AboutBox(self)

    def HelpBox(self):
        webbrowser.open("http://ghmm.org/gql")
        


if __name__ == '__main__':
    app = GQLApp()
    app.mainloop()


